@@ -12,4 +12,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from jax_md import space, energy, minimize, simulate, smap, partition
+from jax_md import space
+from jax_md import energy
+from jax_md import minimize
+from jax_md import simulate
+from jax_md import smap
+from jax_md import partition
+from jax_md import dataclasses
