@@ -13,6 +13,21 @@
 # limitations under the License.
 
 """Code to define different spaces in which particles are simulated.
+
+Spaces are pairs of functions containing:
+
+  * `displacement_fn(Ra, Rb, **kwargs)`
+    Computes displacements between pairs of particles. Ra and Rb should
+    be ndarrays of shape [spatial_dim]. Returns an ndarray of shape
+    [spatial_dim]. To compute displacements over sets of positions, use
+    vmap. Soon (TODO) we will add convenience functions to do this where
+    needed.
+
+  * `shift_fn(R, dR, **kwargs)` Moves points at position R by an amount dR.
+
+In each case, **kwargs is optional keyword arguments that can be supplied to
+the different functions. In cases where the space features time dependence
+this will be passed through a "t" keyword argument.
 """
 
 from __future__ import absolute_import
@@ -23,7 +38,7 @@ from jax.abstract_arrays import ShapedArray
 
 from jax import eval_shape
 from jax import vmap
-from jax import custom_transforms
+from jax import custom_jvp
 
 import jax
 
@@ -62,54 +77,9 @@ def _small_inverse(T):
   # TODO(schsam): Check whether matrices are singular. @ErrorChecking
   return np.linalg.inv(T)
 
-# Tempororay code to do transforms while JAX's custom_transform + vmap is
-# broken. See TODO below.
-from jax.interpreters import ad
-from jax.interpreters import batching
-from jax.abstract_arrays import raise_to_shaped
-from jax import lax
-from jax.api import _dtype
-
-def transform_shape_rule(T, v):
-  return v.shape
-def transform_dtype_rule(T, v):
-  return v.dtype
-def transform_translation_rule(c, T, v):
-  v_dim = len(c.GetShape(v).dimensions()) - 1
-  return c.DotGeneral(v, T, (((v_dim,), (0,)), ((), ())))
-transform_p = lax.standard_primitive(
-  transform_shape_rule, transform_dtype_rule,
-  'transform', transform_translation_rule)
-def transform_batching_rule(operands, batch_dims):
-  T, v = operands
-  T_dim, v_dim = batch_dims
-
-  assert T_dim is None
-  assert v_dim == 0
-
-  return transform_p.bind(T, v), v_dim
-batching.primitive_batchers[transform_p] = transform_batching_rule
-def transform_jvp(primals, tangents):
-  T, v = primals
-  gT, gv = tangents
-  return transform_p.bind(T, v), gv
-ad.primitive_jvps[transform_p] = transform_jvp
 
+@custom_jvp
 def transform(T, v):
-  T_dtype, v_dtype = _dtype(T), _dtype(v)
-  if T_dtype != v_dtype:
-    higher_dtype = lax.dtypes.promote_types(T_dtype, v_dtype)
-    if higher_dtype == v_dtype:
-      T = lax.convert_element_type(T, v_dtype)
-    else:
-      v = lax.convert_element_type(v, T_dtype)
-  return transform_p.bind(T, v)
-
-
-# TODO(schsam): Replace by a custom JVP rule rather than a full primitive
-# when JAX fixes their custom_transforms + vmap issues.
-@custom_transforms
-def _transform(T, v):
   """Apply a linear transformation, T, to a collection of vectors, v.
 
   Transform is written such that it acts as the identity during gradient
@@ -124,7 +94,13 @@ def _transform(T, v):
   """
   _check_transform_shapes(T, v)
   return np.dot(v, T)
-jax.defjvp(_transform, None, lambda g, ans, T, v: g)
+
+
+@transform.defjvp
+def transform_jvp(primals, tangents):
+  T, v = primals
+  dT, dv = tangents
+  return transform(T, v), dv
 
 
 def pairwise_displacement(Ra, Rb):
@@ -193,25 +169,7 @@ def periodic_shift(side, R, dR):
   return np.mod(R + dR, side)
 
 
-"""
-Spaces
-
-  The following functions provide the necessary transformations to perform
-  simulations in different spaces.
-
-  Spaces are tuples containing:
-      displacement_fn(Ra, Rb, **kwargs): Computes displacements between pairs of
-        particles. Ra and Rb should be ndarrays of shape [spatial_dim]. Returns
-        an ndarray of shape [spatial_dim]. To compute displacements over sets
-        of positions, use vmap. Soon (TODO) we will add convenience functions
-        to do this where needed.
-
-      shift_fn(R, dR, **kwargs): Moves points at position R by an amount dR.
-
-  In each case, **kwargs is optional keyword arguments that can be supplied to
-  the different functions. In cases where the space features time dependence
-  this will be passed through a "t" keyword argument.
-"""
+""" Spaces """
 
 
 def free():
